import sys
import math as m
import numpy as np
from numba import jit
from numba.typed import List
import matplotlib.pyplot as plt
import periodictable
from lmfit import minimize, Parameters, Parameter, report_fit, fit_report
from scipy.special import erf

e_radius = periodictable.constants.electron_radius*1e10
energy = 20 # keV


#### data generating ####

def sldCalFun(d,rho,sigma,x):

    N = len(rho) - 1 # number of internal interfaces
    z = [0] + [sum(d[:j+1]) for j in range(len(d))] # height of each interface
    sld_tot = np.zeros(x.shape)
    
    # sld calculation model in Wei's paper 
    for i in range(N):
        erfx = (x-z[i])/sigma[i]/np.sqrt(2)
        sld = 0.5 * (erf(erfx) * (rho[i+1]-rho[i]))
        sld_tot = sld_tot + sld
    sld_tot = sld_tot + 0.5 * (rho[0]+rho[N])

    return sld_tot

def refCalFun(d,rho,mu,sigma,x,rrf=False):
    
    if sigma[0] <= 0: sigma[0] = 1e-5 # eliminate zero
    sigma = sigma[0] * np.ones(len(sigma)) # fixed sigma

    erad = e_radius
    slab=0.25

    x_ = List()
    for xx in x: x_.append(xx)

    k0=2*np.pi*float(energy)/12.3984 # wave vector
    lamda=2*np.pi/k0
    theta=x/2/k0   # convert q to theta
   
    # total length of inner slabs plus 4 times roughness for both sides
    length = np.sum(d) + 4* (sigma[0]+sigma[-1])
    steps=int(length/slab) # each sliced box has thickness of ~ 0.25 A
    xsld=np.linspace(-4*sigma[0],np.sum(d)+4*sigma[-1],steps) # get the z-axis for sld

    sd=length/steps # thickness for each slab
    intrho=sldCalFun(d,rho,sigma,xsld) # rho for all the steps
    intmu=sldCalFun(d,mu,sigma,xsld) # mu for all the steps
    

    # was....
    d, sdel, sbet = List(), List(), List()
    sdel.append(float(rho[0])) # delta for the top phase
    sbet.append(float(mu[0]/2/k0/1e8))        # beta for the top phase
    for rho_ in intrho: # add rho for the interface
        sdel.append(float(rho_))
    sdel.append(float(rho[-1]))  # delta for the bottom phase
    # add delta for the interface
    for mu_ in intmu: # add beta for the interface
        sbet.append(float(mu_/2/k0/1e8))
    sbet.append(float(mu[-1]/2/k0/1e8))    # beta for the bottom phase         
    for i in sdel: d.append(slab)
    ref,refr=parratt(x_, lamda, d, sdel, sbet)

    d_, sdel_, sbet_ = List(), List(), List()
    for i in [0.0,1.0]: d_.append(i)
    for i in [sdel[0],sdel[-1]]: sdel_.append(i)
    for i in [sbet[0],sbet[-1]]: sbet_.append(i)
    frsnll,frsnl1=parratt(x_, lamda,d_, sdel_, sbet_) 

    if rrf == True:
        return ref/frsnll
    else:
        return ref

def ref2min(params,x,y,yerr,fit=True):

    ndata = len(x)

    # allocate parameters to different lists
    sigma_t, rho_b, mu = [], [], []
    d, sigma, rho, qoff = [], [], [], []

    par = params.valuesdict()
    for p in par:
        if p.startswith('sigma'):
            if p.endswith('0'):
                sigma_t.append(par[p]); continue
            else:
                sigma.append(par[p]); continue
        if p.startswith('d'):
            d.append(par[p]); continue
        if p.startswith('mu'):
            mu.append(par[p]); continue
        if p.startswith('qoff'):
            qoff.append(par[p]); continue
        if p.startswith('rho'):
            if '_b' in p:
                rho_b.append(par[p]); continue
            else:
                rho.append(par[p]); continue

    if fit==True: # return residual: 1D array
        residual = np.array([])
        for i in range(ndata):
            yy = refCalFun(d,
                           rho+[rho_b[i]],
                           mu,
                           [sigma[i]]+sigma_t,
                           x[i])
            residual = np.append(residual, (yy-y[i])/yerr[i])             
        return residual
    else: # return model: (y1,y2,y3,y4,y5)
        model = []
        for i in range(ndata):
            yy = refCalFun(d,
                           rho + [rho_b[i]],
                           mu,
                           [sigma[i]] + sigma_t,
                           x[i])
            model.append(yy)
        return tuple(model)
   
def iterCallBack(params,iteration,resid,x,y,err,fit=True):
    if fit== False: return None
    m = sum([params[p].vary for p in params]) # number of parameters
    n = sum([len(xx) for xx in x]) # number of points
    df = n - m # degree of freedom
    redchisq = sum(resid**2) / df
    if (iteration<=10) or (iteration%10==0):  #display reduced chisq every 10 iteration
        print(iteration,redchisq)

def initParameters(name_list,para_list):
    
    if len(name_list)!=len(para_list):
        print(" Does name_list and value_list match, please check ")
        return
    params = Parameters()
    for i,name in enumerate(name_list):
        p = para_list[i]
        params.add(name,value=p[0],vary=p[1],min=p[2],max=p[3])
    return params   

def updateParameters(mrefpar,refparaname,refpara):
    '''Return the parameter name for the items selected in the table.'''
    ui = mrefpar  # it only accepts ref_multiFit_par.ui as ui
    
    parameter_list = refpara
    name_list = refparaname
    index_dict = name2index(name_list,reverse=True)

    # update the value of each cell in the table
    for index in index_dict:
        row, col = index
        value = float(ui.parTW.item(row,col).text())
        i = name_list.index(index_dict[index])
        parameter_list[i][0] = value # update parameter value
        parameter_list[i][1] = False # clear the vary status first

    # update "vary" status in the selected cells
    selected_items = ui.parTW.selectionModel().selectedIndexes()
    selected_names = []
    for item in selected_items:
        selected_index = (item.row(), item.column())
        selected_name = index_dict[selected_index]
        i = name_list.index(selected_name)
        parameter_list[i][1] = True # update vary status
    
    return parameter_list
   
def name2index(name_list, reverse=False):
    ''' Create a mapping from table index to parameter name, or vice versa.
    If reverse=False, create a dict in the form of {name:position}, otherwise
    {position:name}.'''
    ndata = len([p for p in name_list if p.startswith('qoff')])
    nlayer = len([p for p in name_list if p.startswith('d')])
    
    index_dict = {}
    k = 0
    for row in range(ndata+nlayer+1):
        for col in range(4):
            if (row==0) & (col==0): continue
            elif (row>nlayer) & ((col==0)|(col==3)): continue
            else: 
                index_dict[name_list[k]] = (row,col)
                k = k + 1
    if reverse == False:
        return index_dict
    else:
        name_dict = {v:k for k,v in index_dict.iteritems()}
        return name_dict
  
def readData(rrf_files,sel_rows,fit_range,err_type=0):

    '''read multiple data set and cut them to fit range'''
    
    if sel_rows == []: return None
    if fit_range == None: fit_range = [0,1]
    # import pdb; pdb.set_trace();
    rrf = [np.loadtxt(rrf_files[r],comments='#') for r in sel_rows] 
    for i,d in enumerate(rrf):
        select = (d[:,0]>=fit_range[0])&(d[:,0]<=fit_range[1])
        rrf[i] = d[select]
    qz = tuple([a[:,0] for a in rrf]) # tuple: (qz1,qz2,...)
    y = tuple([a[:,1] for a in rrf]) # tuple: (data1,data2,...)
    if err_type==0: # tuple: (err1,err2,...)
        yerr = tuple([a[:,2] for a in rrf])
    elif err_type==1:
        yerr = tuple([np.sqrt(a[:,1]) for a in rrf])
    elif err_type==2:
        yerr = tuple([a[:,1] for a in rrf])
    else:
        yerr=tuple([np.ones(a[:,0].shape) for a in rrf])
       
    return (qz, y, yerr)
    
@jit(nopython=True)
def parratt(q, lamda, d, rho, beta):
    """
        Calculation of reflectivity by Parrat Recursion Formula without any roughness.
        Directly translated from a fortran subroutine 'parratt' in xr_ref.f90, which is
    attached as following:

    subroutine parratt(q,lambda,d,rho,beta,Rgen,Rgenr,M,N)
    !***************************************************************************
    !Calculation of reflectivity by Parratt Recursion Formula without any roughness
    !
    !M = No. of data points
    !N = No. of slabs
    !lambda = wavelength
    !d = list of thicknesses of each slab
    !rho=list of Electron densities of each slab
    !beta=list of Absorption coefficient in each slab
    !Rgen = generated reflectivtiy data
    !Rgenr= generated reflectance data
    !q = change in wave vector
    !***************************************************************************
    integer :: M,N
    double precision :: q(0:M), Rgen(0:M)
    double precision :: d(0:N+1), rho(0:N+1), beta(0:N+1), qc2(0:N+1)
    double precision :: lambda
    double complex :: X, fact1, fact2, r(0:N+1), k1, k2, fact,Rgenr(0:M)
    double precision, parameter :: re=2.817938e-5, pi=3.14159

    do j=0,N+1
       qc2(j)=16.0d0*pi*re*(rho(j)-rho(0))
    enddo

    do i = 0,M
       r(N+1)=dcmplx(0.0d0,0.0d0)
       do j=N,0,-1
          k1=cdsqrt(dcmplx(q(i)**2-qc2(j),-32.0d0*beta(j)*pi**2/lambda**2))
          k2=cdsqrt(dcmplx(q(i)**2-qc2(j+1),-32.0d0*beta(j+1)*pi**2/lambda**2))
          X=(k1-k2)/(k1+k2)
          fact1=dcmplx(dcos(dble(k2)*d(j+1)),dsin(dble(k2)*d(j+1)))
          fact2=dexp(-aimag(k2)*d(j+1))
          fact=fact1*fact2
          r(j)=(X+r(j+1)*fact)/(1.0+X*r(j+1)*fact)
       enddo
       Rgenr(i)=r(0)
       Rgen(i)=cdabs(r(0))**2
    enddo   
    end subroutine parratt

    """
    M = len(q) - 1
    N = len(d) - 2
    r = np.ones(N+2) * (0+0j) # this definition is compatible to numba
    qc2 = np.zeros(N+2)
    Rgen = np.zeros(M+1)
    Rgenr = np.ones(M+1) * (0+0j)

    for j in range(N+2):
        qc2[j] = 16.0 * np.pi * e_radius * (rho[j] - rho[0])
    

    for i in range(M+1):
        r[N+1] = 0+0j
        for j in range(N,-1,-1):
           k1 = np.sqrt(q[i]**2-qc2[j] - 32.0*beta[j]*np.pi**2/lamda**2*1j)
           k2 = np.sqrt(q[i]**2-qc2[j+1] - 32.0*beta[j+1]*np.pi**2/lamda**2*1j)
           X = (k1 - k2) / (k1 + k2)
           fac1 = m.cos(k2.real*d[j+1]) + m.sin(k2.real*d[j+1])*1j
           fac2 = m.exp(-k2.imag*d[j+1])
           fact = fac1 * fac2
           r[j] = (X + r[j+1]*fact) / (1.0+X*r[j+1]*fact)
        Rgenr[i] = r[0]
        Rgen[i] = abs(r[0])**2
    return (Rgen, Rgenr) 





################################################################################
################################################################################

if __name__ == '__main__':
    rho_t = 0.25913738441154344
    rho_b = 0.337
    itMu = 2.792661024598891e-09
    ibMu = 7.0179999999999995e-09
    qz = np.array([0.00879463, 0.00878271, 0.00877078, 0.00875884, 0.00874688,
       0.0087349 , 0.00872291, 0.0087109 , 0.00869887, 0.00868683,
       0.00867477, 0.00866269, 0.0086506 , 0.00863849, 0.00862636,
       0.00861422, 0.00860205, 0.00858987, 0.00857768, 0.00856546,
       0.00855323, 0.00854098, 0.00852872, 0.00851643, 0.00850413,
       0.00849181, 0.00847948, 0.00846712, 0.00845475, 0.00844236,
       0.00842995, 0.00841752, 0.00840507, 0.00839261, 0.00838012,
       0.00836762, 0.0083551 , 0.00834256, 0.00833   , 0.00831742,
       0.00830483, 0.00829221, 0.00827958, 0.00826692, 0.00825425,
       0.00824156, 0.00822884, 0.00821611, 0.00820336, 0.00819059,
       0.0081778 , 0.00816498, 0.00815215, 0.0081393 , 0.00812643,
       0.00811353, 0.00810062, 0.00808769, 0.00807473, 0.00806176,
       0.00804876, 0.00803574, 0.0080227 , 0.00800964, 0.00799656,
       0.00798346, 0.00797033, 0.00795719, 0.00794402, 0.00793083,
       0.00791762, 0.00790439, 0.00789113, 0.00787785, 0.00786455,
       0.00785123, 0.00783788, 0.00782452, 0.00781112, 0.00779771,
       0.00778427, 0.00777081, 0.00775733, 0.00774382, 0.00773029,
       0.00771673, 0.00770316, 0.00768955, 0.00767593, 0.00766227,
       0.0076486 , 0.0076349 , 0.00762118, 0.00760743, 0.00759365,
       0.00757985, 0.00756603, 0.00755218, 0.0075383 , 0.0075244 ,
       0.00751048, 0.00749652, 0.00748255, 0.00746854, 0.00745451,
       0.00744045, 0.00742637, 0.00741226, 0.00739812, 0.00738396,
       0.00736977, 0.00735555, 0.0073413 , 0.00732703, 0.00731272,
       0.00729839, 0.00728403, 0.00726965, 0.00725523, 0.00724079,
       0.00722631, 0.00721181, 0.00719728, 0.00718272, 0.00716813,
       0.00715351, 0.00713886, 0.00712418, 0.00710947, 0.00709473])
    ref = refCalFun([], [rho_t, rho_b], [itMu,ibMu], [3.0], qz)
    print(ref)



import sys
import os
# Use absolute path instead of relative path ('./') to avoid trouble when installed by pip
dir_path = os.path.dirname(os.path.realpath(__file__))
print(dir_path)
UI_path = dir_path + '/GUI/'
import time
import multiprocessing


######################################################
# This block of code is needed for properly working with PyInstaller

# import the following three modules in order to work with PyInstaller

# mplwidget is imported explicitly here because PyInstaller needs to find it.

# Define function to import external files when using PyInstaller.
# https://stackoverflow.com/questions/37888581/pyinstaller-ui-files-filenotfounderror-errno-2-no-such-file-or-directory
def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        # when not bundled by PyInstaller, normal method is used.
        base_path = os.path.abspath(".")
    return os.path.join(base_path, relative_path)


#######################################################
from collections import OrderedDict

from PyQt5 import uic
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

from scipy.interpolate import interp1d
from scipy.special import *
import numpy as np
import matplotlib.pyplot as plt
from cycler import cycler
default_cycler = (cycler(color=['b','g','c','m','y','k']))
plt.rc('axes', prop_cycle=default_cycler)
plt.rc('lines',markersize=3, linewidth=1)

import lmfit as lm # fitting module
import periodictable as pdtb # a module for periodic table
r_e = pdtb.constants.electron_radius * 1e10  # classical electron radius, in A
N_A = pdtb.constants.avogadro_number  # Avogadro number, unitless
k_B = 1.38065e-23  # Boltzman constant, in J/K

# user defined module
import flu_routines_new as fl

# Here the absolute path is used because PyInstaller needs to find it.
(Ui_MainWindow, QMainWindow) = uic.loadUiType(resource_path(UI_path + 'mainwindow.ui'))


class myThread(QThread):
    def __init__(self,func):
        QThread.__init__(self)
        self.func = func

    def run(self):
        self.func
        
def fluErrorFitSingleCore2(i, value_list, sh, qz, pname, flu_par, flucal_par, data_to_fit):
        flu_par[pname].value = value_list[i]  # change value of the chosen parameter
        fluerr_result = lm.minimize(fl.flu2min, flu_par,
                                    args=((sh, qz), flucal_par),
                                    kws={'data': data_to_fit[:, 1], 'eps': data_to_fit[:, 2]})
        return ([i, value_list[i], fluerr_result.nfree, fluerr_result.redchi])

def multiCore(func, iterable):
    pool = multiprocessing.Pool()
    result = pool.map(func, range(len(iterable)))
    pool.close()
    pool.join()
    return result


class MainWindow (QMainWindow):
    """MainWindow inherits QMainWindow"""

    def __init__(self, parent = None):
        QMainWindow.__init__(self, parent)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.ui.tabWidget.setCurrentIndex(1)
        self.directory=os.getcwd()

        self.halftab = '    '
        self.flusavefitindex = 0
        self.flufiles = []
        self.flufitfiles = []
        self.fludata = []
        self.flufitdata = []
        self.beam = 'uniform'
        self.flucal_par = OrderedDict()
        self.flu = 0
        self.selectedflufiles_rows = []
        self.selectedflufitfiles_rows = []
        self.eleradius = pdtb.constants.electron_radius*1e10
        self.avoganum = pdtb.constants.avogadro_number
        self.boltzmann = 1.38065e-23
        self.errorlist = np.array([[1, 1.074], [2, 1.204], [3, 1.222],
                                [4, 1.220], [5, 1.213], [6, 1.205], 
                                [7, 1.198], [8, 1.191], [9, 1.184], 
                                [10, 1.178], [11, 1.173], [12, 1.168], 
                                [13, 1.163], [14, 1.159], [15, 1.155], 
                                [16, 1.151], [17, 1.148], [18, 1.145], 
                                [19, 1.142], [20, 1.139], [22, 1.134], 
                                [24, 1.129], [26, 1.125], [28, 1.121], 
                                [30, 1.118], [32, 1.115], [34, 1.112], 
                                [36, 1.109], [38, 1.106], [40, 1.104], 
                                [42, 1.102], [44, 1.100], [46, 1.098], 
                                [48, 1.096], [50, 1.094], [60, 1.087], 
                                [70, 1.081], [80, 1.076], [90, 1.072], 
                                [100, 1.069], [120, 1.063], [140, 1.059], 
                                [160, 1.055], [180, 1.052]]) #, [3000, 1.050]])

        self.setupUI()
        self.updatePar()
        self.debugErr()

    def setupUI(self):
        self.ui.addflufilePB.clicked.connect(self.addFluFile)
        self.ui.flufileLW.itemSelectionChanged.connect(self.updateSelectedFluFile)
        self.ui.rmflufilePB.clicked.connect(self.removeFluFile)
        self.ui.addflufitfilePB.clicked.connect(self.addFluFitFile)
        self.ui.flufitfileLW.itemSelectionChanged.connect(self.updateSelectedFluFitFile)
        self.ui.rmflufitfilePB.clicked.connect(self.removeFluFitFile)
        self.ui.fluxaxisCB.currentIndexChanged.connect(self.updateUI)
        self.ui.fluqcCB.stateChanged.connect(self.updateFluPlot)
        self.ui.flulineCB.stateChanged.connect(self.updateFluPlot)
        self.ui.flulegendCB.stateChanged.connect(self.updateFluPlot)
        self.ui.flulegendlocCoB.currentIndexChanged.connect(self.updateFluPlot)
        self.ui.flulogyCB.stateChanged.connect(self.updateFluPlot)
        self.ui.flugridCB.stateChanged.connect(self.updateFluPlot)
        self.ui.flushowCB.stateChanged.connect(self.updateFluPlot)
        self.ui.flucompCB.stateChanged.connect(self.updateFluPlot)
        self.ui.flusimuPB.clicked.connect(self.updateFluCal)
        self.ui.flufitPB.clicked.connect(self.fitFlu)
        self.ui.flusaveCB.activated.connect(self.saveFlu)
        self.ui.fluloadCB.activated.connect(self.loadFlu)
        self.ui.insflusubPB.clicked.connect(self.insFluIon)
        self.ui.rmflusubPB.clicked.connect(self.rmFluIon)
        self.ui.fluErrPB.clicked.connect(self.fluErrorInit)
        self.ui.flulimitPB.clicked.connect(self.setupLimitsUI)
        # connect system parameter signals
        self.ui_syspar = OrderedDict(
            [('E_inc',self.ui.fluIncEnLE),
             ('E_emt',self.ui.fluEmtEnLE),
             ('mu_top_inc',self.ui.flumutopincLE),
             ('mu_top_emt', self.ui.flumutopemtLE),
             ('mu_bot_inc', self.ui.flumubotincLE),
             ('mu_bot_emt',self.ui.flumubotemtLE),
             ('rho_top', self.ui.flurhotopLE),
             ('rho_bot', self.ui.flurhobotLE),
             ('width', self.ui.fluwidthLE),
             ('det_len',self.ui.fludetLE)]
        )
        for p,u in self.ui_syspar.items():
            u.returnPressed.connect(self.updatePar)
            u.returnPressed.connect(self.updateFluCal)
        self.ui.flubmpfCombo.currentIndexChanged.connect(self.updatePar)

        # connect fitting parameter signals
        self.ui_params = OrderedDict(
            [('losc', [self.ui.flubotscaleLE, self.ui.flubotscaleCB]),
             ('hisc', [self.ui.flutopscaleLE, self.ui.flutopscaleCB]),
             ('lobk', [self.ui.flubotbulkLE, self.ui.flubotbulkCB]),
             ('upbk', [self.ui.flutopbulkLE, self.ui.flutopbulkCB]),
             ('surd', [self.ui.flusurdLE, self.ui.flusurdCB]),
             ('bg',   [self.ui.flubgLE, self.ui.flubgCB]),
             ('qoff', [self.ui.fluqoffLE, self.ui.fluqoffCB]),
             ('curv', [self.ui.flucurvLE, self.ui.flucurvCB]),
             ('loff', [self.ui.fluloffLE, self.ui.fluloffCB]),
             ('soff', [self.ui.flusoffLE, self.ui.flusoffCB])]
        )
        for p,u in self.ui_params.items():
            u[0].returnPressed.connect(self.updatePar)
            u[0].returnPressed.connect(self.updateFluCal)
            u[1].stateChanged.connect(self.updatePar)

        # set up a container for parameter limits
        self.par_limits = OrderedDict()
        for p,u in self.ui_params.items():
            if p in ['lobk','upbk','surd','bg','curv']:
                self.par_limits[p] = [0, 1000]
            elif p in ['hisc','losc']:
                self.par_limits[p] = [1e-10, 1000]
            else:
                self.par_limits[p] = [None, None]

    def updateUI(self,index=None, fresh=True):
        '''
        This function also works as a slot for qcombobox signal, so the first arg/kws by default
        is reserved for receiving signal argument, which can be either index or text.
        '''
        print("{}, {}".format(index,fresh))
        # set text for fitting parameters
        for p, u in self.ui_params.items():
            if p in ['curv', 'loff']:
                u[0].setText(format(self.flu_par[p].value, '.4f'))
            else:
                u[0].setText(format(self.flu_par[p].value, '.2e'))

        # set text for system parameters
        for p, u in self.ui_syspar.items():
            u.setText(format(self.sys_par[p], '.4f'))

        # set beam profile
        index = self.ui.flubmpfCombo.findText(self.beam)
        self.ui.flubmpfCombo.setCurrentIndex(index)
        self.xaxis = self.ui.fluxaxisCB.currentText()

        if fresh == True:
            if self.xaxis == 'Qz':
                _xrange = '0.005:0.016'
                self.ui.fluloffCB.setCheckable(True)
                self.ui.fluloffCB.setText('L2 offset')
                self.ui.fluloffqzlabel.setText('mm')
            elif self.xaxis == 'Sh':
                _xrange = '-0.1:0.1'
                self.ui.fluloffCB.setText('Qz')
                self.ui.fluloffLE.setText(format(0.006, '.4f'))
                self.ui.fluloffqzlabel.setText(self.ui.fluqofflabel.text())
            self.ui.fluSimuRangeLE.setText(_xrange)
            self.ui.fluFitRangeLE.setText(_xrange)
        else:
            try:
                _simu_range = ':'.join(str(self.flu_simu_range)[1:-1].split(','))
                _fit_range = ':'.join(str(self.flu_fit_range)[1:-1].split(','))
                self.ui.fluSimuRangeLE.setText(_simu_range)
                self.ui.fluFitRangeLE.setText(_fit_range)
                self.ui.fluloffLE.setText(str(self.qz[0]))
            except Exception as e:
                self.updateUI(fresh=True)

        self.updatePar()

    def updatePar(self):  #initialize the flu parameters

        # system parameters for fluorescence
        self.beam = str(self.ui.flubmpfCombo.currentText()) # beam profile
        self.sys_par = OrderedDict()
        for p, u in self.ui_syspar.items():
            self.sys_par[p] = float(u.text())
        self.sys_par['beam'] = self.ui.flubmpfCombo.currentText()
        self.sys_par['span'] = 75.6 # the length of sample cell, in mm.

        # fitting parameters for fluorescence
        self.flu_par = lm.Parameters()
        try:
            for name,par in self.ui_params.items():
                self.flu_par.add(name, value=float(par[0].text()), vary=par[1].isChecked(),
                                       min=self.par_limits[name][0], max=self.par_limits[name][1],
                                       expr=None, brute_step=None)
        except ValueError as VE:
            print("ValueError: ", VE)
        else:
            # fitting (if any) parameters for reflectivity
            self.ref_par = lm.Parameters()
                # add tuples:       (NAME       VALUE   VARY MIN  MAX  EXPR BRUTE_STEP)
            self.ref_par.add_many( ('rho_t', self.sys_par['rho_top'], 0, None, None, None, None),
                                   ('rho_b', self.sys_par['rho_bot'], 0, None, None, None, None),
                                   ('mu_t', self.sys_par['mu_top_inc'], 0, None, None, None, None),
                                   ('mu_b', self.sys_par['mu_bot_inc'], 0, None, None, None, None),
                                   ('sigma0', 3.0, 0, None, None, None, None),
                                   ('q_off', 0, 0, None, None, None, None ))
            # info of element in the system
            self.flu_elements = [['Eu', 1, 0.947]]  # name, composition, Ionic Radius(A)

        # update the list of sh and qz
        try:
            self.flu_simu_range = [float(i) for i in str(self.ui.fluSimuRangeLE.text()).split(':')]
            self.flu_fit_range = [float(i) for i in str(self.ui.fluFitRangeLE.text()).split(':')]
        except:
            print("Error: Check if the range is right.")
        self.xaxis = self.ui.fluxaxisCB.currentText()
        if self.xaxis == 'Qz':
            self.sh = np.array([0])
            self.qz = np.linspace(self.flu_simu_range[0], self.flu_simu_range[1], 200)
        elif self.xaxis == 'Sh':
            self.qz = np.array([float(self.ui.fluloffLE.text())])
            self.sh = np.linspace(self.flu_simu_range[0], self.flu_simu_range[1], 200)

        # update parameters for flurescence calculation
        self.flucal_par = fl.update_flu_parameters(self.flucal_par,
                                                   self.flu_par,
                                                   self.sys_par,
                                                   self.flu_elements)

    def updateLimits(self):
        ui = self.ui_limits
        try:
            for p, u in self.ui_par_limits.items():
                if u[0].isChecked():
                    try:
                        self.par_limits[p][0] = float(u[1].text())
                        self.par_limits[p][1] = float(u[2].text())
                        assert self.par_limits[p][1] >= self.par_limits[p][0]
                    except ValueError as e:
                        print("{} Please provide a valid limit for '{}'. ".format(e,u[0].text()))
                        raise
                    except AssertionError:
                        print("Max should be larger than Min for {}".format(u[0].text()))
                        raise
                else:
                    self.par_limits[p][0] = None
                    self.par_limits[p][1] = None
        except:
            print("An Error occurs... See above")
        else:
            self.updatePar()
            ui.close()

    def setupLimitsUI(self):

        ui = uic.loadUi(UI_path + 'err4.ui', QDialog(self))
        ui.cancelPB.clicked.connect(ui.close)
        ui.confirmPB.clicked.connect(self.updateLimits)
        self.ui_par_limits = OrderedDict(
            [('losc', [ui.flubotscaleCB, ui.minbotscaleLE, ui.maxbotscaleLE]),
             ('hisc', [ui.flutopscaleCB, ui.mintopscaleLE, ui.maxtopscaleLE]),
             ('lobk', [ui.flubotbulkCB, ui.minbotbulkLE, ui.maxbotbulkLE]),
             ('upbk', [ui.flutopbulkCB, ui.mintopbulkLE, ui.maxtopbulkLE]),
             ('surd', [ui.flusurdCB, ui.minsurdLE, ui.maxsurdLE]),
             ('bg', [ui.flubgCB, ui.minbgLE, ui.maxbgLE]),
             ('qoff', [ui.fluqoffCB, ui.minqoffLE, ui.maxqoffLE]),
             ('curv', [ui.flucurvCB, ui.mincurvLE, ui.maxcurvLE]),
             ('loff', [ui.fluloffCB, ui.minloffLE, ui.maxloffLE]),
             ('soff', [ui.flusoffCB, ui.minsoffLE, ui.maxsoffLE])]
        )
        for p, u in self.ui_par_limits.items():
            if self.par_limits[p][0] is None and self.par_limits[p][1] is None:
                u[0].setChecked(False)
                continue
            else:
                u[0].setChecked(True)
            try:
                u[1].setText(format(self.par_limits[p][0], '.2e'))
            except TypeError:
                pass  # leave empty if limit is None
            try:
                u[2].setText(format(self.par_limits[p][1], '.2e'))
            except TypeError:
                pass
        self.ui_limits = ui
        ui.show()


    def addFluFile(self): #add flu files into the listwidget and deselect all flu files in the listwidget

        f, _ = QFileDialog.getOpenFileNames(
            caption='Select Multiple Fluorescence Files to import',
            directory=self.directory,
            filter='Flu Files (*.flu*;*_flu.txt)'
        )
        self.flufiles = self.flufiles + f
        self.directory = str(QFileInfo(self.flufiles[0]).absolutePath())
        self.updateFluFile()

    def updateFluFile(self): # update flu files in the listwidget
        self.ui.flufileLW.clear()
        for i, f in enumerate(self.flufiles):
            try:
                self.ui.flufileLW.addItem('#'+str(i+1)+self.halftab+str(f.split('\\')[-2])+'\\'+str(f.split('\\')[-1]))
            except:
                self.ui.flufileLW.addItem('#'+str(i+1)+self.halftab+str(f.split('/')[-2])+'/'+str(f.split('/')[-1]))

    def updateSelectedFluFile(self): #update the selected flu files in the listwidget
        self.fludata = []
        selectedflufiles = self.ui.flufileLW.selectedItems()
        self.selectedflufiles_rows = [self.ui.flufileLW.row(item) for item in selectedflufiles]
        self.selectedflufiles_rows.sort()
        if len(selectedflufiles) != 0:
            try:
                for i, r in enumerate(self.selectedflufiles_rows):
                    data = np.loadtxt(str(self.flufiles[r]),comments='#')
                    for d in data: # replace zero error bar with 10% error
                        if d[2]==0:
                            print('Error bar replaced with 10% of value for entry {}'.format(d))
                            d[2] = float(d[1]) / 10
                    print('\n')

                    self.fludata.append(data)
            except OSError as e:
                print(e)
        self.updateFluPlot()

    def removeFluFile(self): #remove flu files in the listwidget and deselect all flu files in the listwidget

        to_del = [self.ui.flufileLW.row(item) for item in self.ui.flufileLW.selectedItems()]
        self.flufiles = [f for i,f in enumerate(self.flufiles) if i not in to_del]

        self.ui.flufileLW.clear()
        self.updateFluFile()

    def addFluFitFile(self): #add flu fit files into the listwidget and deselect flu fit files in the listwidget
        try:
            f, _ = QFileDialog.getOpenFileNames(
                caption = 'Select Multiple Fluorescence Fit Files to import',
                directory = self.directory,
                filter = 'FIT Files (*.fit*; *_fit.txt)'
            )
            self.flufitfiles = self.flufitfiles + f
            self.directory = str(QFileInfo(self.flufitfiles[0]).absolutePath())
            self.updateFluFitFile()
        except IndexError as IE:
            pass # ignore IndexError
        except:
            print("Something went wrong when reading fit files!")
    def updateFluFitFile(self): #update flu fit files in the listwidget
        self.ui.flufitfileLW.clear()
        for i, f in enumerate(self.flufitfiles):
            try:
                self.ui.flufitfileLW.addItem(
                    '#' + str(i + 1) + self.halftab + str(f.split('\\')[-2]) + '\\' + str(f.split('\\')[-1]))
            except:
                self.ui.flufitfileLW.addItem(
                    '#' + str(i + 1) + self.halftab + str(f.split('/')[-2]) + '/' + str(f.split('/')[-1]))

    def updateSelectedFluFitFile(self): #update the selected flu fit files in the listwidget
        self.flufitdata = []
        selectedflufitfiles = self.ui.flufitfileLW.selectedItems()
        self.selectedflufitfiles_rows = [self.ui.flufitfileLW.row(item) for item in selectedflufitfiles]
        self.selectedflufitfiles_rows.sort()
        try:
            if len(selectedflufitfiles) != 0:
                for i, r in enumerate(self.selectedflufitfiles_rows):
                    data = np.loadtxt(str(self.flufitfiles[r]),comments='#')
                    self.flufitdata.append(data)
        except OSError as OE:
            print(OE)
        self.updateFluPlot()

    def removeFluFitFile(self):  #remove flu fit files in the listwidget and deselect all flu fit files in the listwidget

        to_del = [self.ui.flufitfileLW.row(item) for item in self.ui.flufitfileLW.selectedItems()]
        self.flufitfiles = [f for i, f in enumerate(self.flufitfiles) if i not in to_del]

        self.ui.flufitfileLW.clear()
        self.updateFluFitFile()

    def updateFluPlot(self): #update the plot in the flu plotwidget

        ax1 = self.ui.fluPW.canvas.ax
        ax1.clear()

        if self.ui.flulineCB.isChecked():
            ls = '-'
        else:
            ls = ''

        if len(self.fludata) != 0: #plot flu files
            for i,d in enumerate(self.fludata):
                ax1.errorbar(d[:,0],d[:,1],yerr=d[:,2],
                             marker='o', ls=ls,
                             label='#'+str(i+1))

        if len(self.flufitdata) != 0: #plot flu fit files
            for i, d in enumerate(self.flufitdata):
                ax1.plot(d[:, 0], d[:, 1],marker='', ls='-', label=' fit #'+str(i + 1))


        self.xaxis = self.ui.fluxaxisCB.currentText()
        if self.ui.flushowCB.isChecked():
            if np.all(self.flu==0):
                print('Please print simulate button first!!')
                return
            else:
                if self.xaxis == 'Qz':
                    x = self.qz
                    y = self.flu[0, :, 2:]
                    x_range = [x[0]-0.001, x[-1]+0.001]
                    x_label = r'$Q_z$' + ' ' + r'$[\AA^{-1}]$'
                    if self.ui.fluqcCB.isChecked():
                        ax1.axvline(self.qc, color='black', alpha=0.5)
                elif self.xaxis == 'Sh':
                    x = self.sh
                    y = self.flu[:, 0, 2:]
                    x_range = [x[0]-0.001, x[-1]+0.001]
                    x_label = r'$\Delta sh$' + ' ' + r'$[mm]$'
            try:
                ax1.set_xlabel(x_label)
                ax1.set_ylabel(r'$Intensity [a.u.]$')
                ax1.set_xlim(x_range)
                ax1.plot(x, y[:,0], ls='-', label='total', color='r')
                if self.ui.flucompCB.isChecked():
                    ax1.plot(x, y[:,1], ls='-', label='water',color='b', alpha=0.5)
                    ax1.plot(x, y[:,2], ls='-', label='interface',color='purple', alpha=0.5)
                    ax1.plot(x, y[:,3], ls='-', label='oil', color='g', alpha=0.5)

            except ValueError as VE:
                print(VE)
        if self.ui.flulegendCB.isChecked():
            ax1.legend(loc = str(self.ui.flulegendlocCoB.currentText()),
                       frameon=False,
                       scatterpoints=0,
                       numpoints=1)
        if self.ui.flugridCB.isChecked():
            ax1.grid(1)
        if self.ui.flulogyCB.isChecked():
            ax1.set_yscale('log')
        else:
            ax1.set_yscale('linear')

        self.ui.fluPW.canvas.draw()

    def setFluPlotScale(self): #set the scale of each data in the flu plot
        if len(self.selectedflufiles_rows)+len(self.selectedflufitfiles_rows)==0:
            self.messageBox('Warning:: No Fluorescence or Fit files selected!')
        else:
            row_flu=len(self.selectedflufiles_rows)
            row_fit=len(self.selectedflufitfiles_rows)
            row=row_flu+row_fit
            Dialog=QDialog(self)
            self.uiplotscale=uic.loadUi(UI_path + 'plotscale.ui', Dialog)
            self.uiplotscale.scaleTW.setRowCount(row) #set the table size; 4 column is fixed
            self.uiplotscale.show()
            self.uiplotscale.scaleLabel.setText('Fluorescence Plot Scale Setup: X=X*Factor+Offset')
            self.uiplotscale.scaleTW.setHorizontalHeaderLabels(QStringList()<<"X Factor"<<"X Offset"<<"Y Factor"<<"Y Offset") #set the horizontal header
            vlabel=QStringList() #set the vertical header
            for i in range(row_flu):
                vlabel.append("Flu #"+str(self.selectedflufiles_rows[i]+1))
            for i in range(row_fit):
                vlabel.append("Fit #"+str(self.selectedflufitfiles_rows[i]+1))
            self.uiplotscale.scaleTW.setVerticalHeaderLabels(vlabel)
            for i in range(row_flu):  #set the initial values
                for j in range(4):
                    self.uiplotscale.scaleTW.setItem(i,j,QTableWidgetItem(str(self.fluscale[i][j])))
                    self.uiplotscale.scaleTW.item(i,j).setTextAlignment(Qt.AlignCenter)
            for i in range(row_fit):
                for j in range(4):
                    self.uiplotscale.scaleTW.setItem(i+row_flu,j,QTableWidgetItem(str(self.flufitscale[i][j])))
                    self.uiplotscale.scaleTW.item(i+row_flu,j).setTextAlignment(Qt.AlignCenter)
            self.connect(self.uiplotscale.scaleTW, SIGNAL('cellChanged(int,int)'), self.updateFluPlotScale) #update the flu scale and plot
            self.connect(self.uiplotscale.closePB,SIGNAL('clicked()'), self.closePlotScale) #close the scale setup window

    def updateFluPlotScale(self): #update the scale of each data in the flu plot
        row_flu=len(self.selectedflufiles_rows)
        row_fit=len(self.selectedflufitfiles_rows)
        self.fluscale=[[float(str(self.uiplotscale.scaleTW.item(i,j).text())) for j in range(4)] for i in range(row_flu)]
        self.flufitscale=[[float(str(self.uiplotscale.scaleTW.item(i+row_flu,j).text())) for j in range(4)] for i in range(row_fit)]
        self.updateFluPlot()

    def updateFluCal(self): # calculate the flu  based on current parameters.

        self.updatePar()
        # p is the parameter set taken by the core function, must be initialized before updatding
        p = OrderedDict()
        p = fl.update_flu_parameters(p, self.flu_par, self.sys_par,self.flu_elements)
        if not self.ui.flushowCB.isChecked():
            return # if show is not checked, do nothing.
        if self.xaxis == 'Qz':
            self.qc = np.sqrt(2*(p['ibDt']-p['itDt'])) * 2 * p['k0']
        self.flu = fl.flu2min(self.flu_par, (self.sh, self.qz), p)

        self.updateFluPlot()

    def fitFlu(self, uncertainty_calculation=False):

        selectedflufiles = self.ui.flufileLW.selectedItems()
        try:
            assert len(selectedflufiles) == 1
        except:
            print("Error: please select one data to fit.")
            return
        data = self.fludata[0]

        self.updatePar()

        self.data_to_fit = data[(data[:,0] >= self.flu_fit_range[0]) * (data[:, 0] <= self.flu_fit_range[1])]

        if self.xaxis == 'Qz':
            self.qz = self.data_to_fit[:,0]
        elif self.xaxis == 'Sh':
            self.sh = self.data_to_fit[:,0]

        # for uncertainty calculation, this function is only used to get preparaed for multicore processing
        if uncertainty_calculation == True: return

        self.flu_result = lm.minimize(fl.flu2min, self.flu_par,
                                      args=((self.sh, self.qz), self.flucal_par),
                                      kws={'data':self.data_to_fit[:,1], 'eps':self.data_to_fit[:,2]}
                                      )
        self.flu_par = self.flu_result.params

        # for uncertainty calculation, the following is not needed

        tb = self.ui.fluparaTB
        tb.clear()
        tb.append(lm.fit_report(self.flu_result))

        self.updateUI(fresh=False)  # it has to be before updateFluCal()
        self.updateFluCal() # self.updateUI() has to be excucated before it reads parameters from GUI

    def saveFlu(self):
        if str(self.ui.flusaveCB.currentText())=='Save Fit':
            self.saveFluFitDig()
        elif str(self.ui.flusaveCB.currentText())=='Save Para':
            self.saveFluPara()

    def saveFluPara(self):

        self.updatePar()

        self.saveFileName = QFileDialog.getSaveFileName(caption='Save Fluorescence Fitting Parameters',
                                                        directory=self.directory,
                                                        filter='Par Files (*.par*;*_par.txt)')
        with open(self.saveFileName[0] + '_par.txt','w') as fid:
            try:
                try:
                    fid.write('Chi_Square\t' + format(self.flu_result.redchi, '.3f') + '\n')  # chisquare
                except:
                    fid.write('Chi_Square\tNA\n')

                fid.write('Fitting_Parameters\n')
                for p, u in self.ui_params.items():
                    fid.write(p + '\t\t' + format(float(u[0].text()),'.3e') + '\n')

                fid.write('\nSystem_Parameters\n')
                fid.write('Beam_Profile\t\t' + self.beam +'\n')
                for p, u in self.ui_syspar.items():
                    fid.write(p + '\t\t' + format(float(u.text()), '.4f') + '\n')

                print("Parameters saved!")

            except:
                print('Oops! Something went wrong, please check your parameters!')

    def loadFlu(self):
        if str(self.ui.fluloadCB.currentText())=='Load Para':
            self.loadFluPara()

    def loadFluPara(self):

        try:
            filename, _ = QFileDialog.getOpenFileName(caption='Select Parameter File to read',
                                                   directory=self.directory,
                                                   filter='Par Files (*.par*;*_par.txt)')
            self.directory = str(QFileInfo(filename).absolutePath())
            with open(str(filename)) as fid:
                fdata=fid.readlines()
        except IOError: # if the dialog is canceled.
            return
        # set ui values with loaded value
        line_num = 0
        line_type = 0 # 0: not a parameter line; 1: fitting parameter line; 2: system parameter line
        while True:
            try:
                line = fdata[line_num].split()
            except IndexError: # end of file
                break
            if line == []:
                line_type = 0
            else:
                if line[0] == 'Fitting_Parameters':
                    line_type = 1
                elif line[0] == 'Beam_Profile':
                    self.beam = line[1]
                    line_type = 2
                elif line_type == 1:
                    self.flu_par[line[0]].value = float(line[1])
                elif line_type == 2:
                    self.sys_par[line[0]] = float(line[1])
            line_num += 1

        # update parameter with new ui values
        self.updateUI(fresh=False)
        self.updateFluCal()

    def saveFluFitDig(self):

        Dialog=QDialog(self)
        self.uiflusavefit = uic.loadUi(resource_path(UI_path+'refsave.ui'), Dialog)
        self.uiflusavefit.label.setText('Save Fluorescence Fit/Calcualtion!')
        try:
            self.uiflusavefit.xminLE.setText(str(self.flu_simu_range[0]))
            self.uiflusavefit.xmaxLE.setText(str(self.flu_simu_range[1]))
        except:
            pass
        self.uiflusavefit.numpointLE.setText(str(200))

        self.uiflusavefit.cancelPB.clicked.connect(self.cancelSaveFluFit)
        self.uiflusavefit.okPB.clicked.connect(self.saveFluFit)

        self.uiflusavefit.show()

    def cancelSaveFluFit(self):
        self.uiflusavefit.close()
        self.flusavefitindex=0

    def saveFluFit(self):

        try:
            self.flusavefitindex = 1
            self.flunp = float(self.uiflusavefit.numpointLE.text())
            self.fluxmin = float(self.uiflusavefit.xminLE.text())
            self.fluxmax = float(self.uiflusavefit.xmaxLE.text())
            assert (self.fluxmin < self.fluxmax), "Maximum smaller than Minimum"

            self.saveFileName = QFileDialog.getSaveFileName(caption='Save Fluorescence Fit Data',
                                                            directory=self.directory)
            fname = self.saveFileName[0] + '_fit.txt'
            if self.xaxis == 'Qz':
                fit_to_save = self.flu[0,:,(1,2)].transpose()
            elif self.xaxis == 'Sh':
                fit_to_save = self.flu[:,0,(0,2)]
            np.savetxt(fname, fit_to_save, fmt='%.4e\t%.4e')

            self.flusavefitindex=0
            self.uiflusavefit.close()
        except AssertionError as AE:
            print("Error: {0}".format(AE))
        except IndexError as IE:
            pass # Ignore IndexError
        except:
            print("An error happens while saving fit file!")

    def debugErr(self):
        parfile = os.path.join(dir_path,
                               'sh_sample03_318_50mMEu(NO3)3_s1h0.2_qz0.0015_par.txt')
        self.updateFluFile()

        with open(str(parfile)) as fid:
            fdata = fid.readlines()
        # set ui values with loaded value
        line_num = 0
        line_type = 0  # 0: not a parameter line; 1: fitting parameter line; 2: system parameter line
        while True:
            try:
                line = fdata[line_num].split()
            except IndexError:  # end of file
                break
            if line == []:
                line_type = 0
            else:
                if line[0] == 'Fitting_Parameters':
                    line_type = 1
                elif line[0] == 'Beam_Profile':
                    self.beam = line[1]
                    line_type = 2
                elif line_type == 1:
                    self.flu_par[line[0]].value = float(line[1])
                elif line_type == 2:
                    self.sys_par[line[0]] = float(line[1])
            line_num += 1
        # update parameter with new ui values
        self.updateUI()
        self.updateFluCal()

    def fluErrorInit(self):

        # choose the parameter for which the chisq is calculated
        self.fluerr_pname = [] # initialize # of the chosen parameters
        try:
            self.fluerr_pname = [p for p,u in self.ui_params.items() if u[1].isChecked()]
            if len(self.fluerr_pname) != 1:
                raise ValueError
            print("Calculating Chi-square for:", *self.fluerr_pname)
        except ValueError:
            print(" Did u pick the right number of parameters to fit?\n\n")
            # if multiple para's r checked, uncheck all and raise error
            for name in self.fluerr_pname:
                self.ui_params[name][1].setChecked(False)
            return

        self.uifluerr1=uic.loadUi(UI_path + 'err1.ui',QDialog(self))
        self.uifluerr1.label.setText('Uncertainty Calculation for Parameter:' + self.fluerr_pname[0])

        best_value = float(self.ui_params[self.fluerr_pname[0]][0].text())
        half_range_to_fit = abs(best_value*0.1)
        # the length of left and right half of range for the chosen values.
        self.uifluerr1.bestvalLE.setText(format(best_value, '.2e'))
        self.uifluerr1.leftLimitLE.setText(  # set left limit
            format(best_value - half_range_to_fit, '.2e'))
        self.uifluerr1.rightLimitLE.setText( # set right limit
            format(best_value + half_range_to_fit, '.2e'))

        self.uifluerr1.numIntervalLE.setText(format(10  ,'d'))

        # connect the pushbutton to next step
        # self.uifluerr1.cancelPB.clicked.connect(lambda x: self.uifluerr1.close())
        self.uifluerr1.cancelPB.clicked.connect(self.uifluerr1.close)
        self.uifluerr1.nextPB.clicked.connect(self.fluErrorPara)
        self.uifluerr1.show()

    def fluErrorPara(self):

        self.uifluerr1.close()
        # calculate a list of values the parameter should take where the chisq is calculated.
        self.fluerr_best_value = float(self.uifluerr1.bestvalLE.text())
        self.fluerr_left_limit = float(self.uifluerr1.leftLimitLE.text())
        self.fluerr_right_limit = float(self.uifluerr1.rightLimitLE.text())
        self.fluerr_num_points = int(self.uifluerr1.numIntervalLE.text())+1
        # append the fittted value for that parameter for displaying that
        # value in the chisq plot as the red dot.
        self.fluerr_fit_range = np.append(self.fluerr_best_value,
                                          np.linspace(self.fluerr_left_limit,
                                                      self.fluerr_right_limit,
                                                      self.fluerr_num_points))
        self.fluerr_chisq_list = np.zeros(self.fluerr_fit_range.shape)

        # automatically toggle the state of fiting and fixed parameters
        for p, u in self.ui_params.items():  u[1].toggle()

        # close the first dialog and open a new dialog
        self.uifluerr2 = uic.loadUi(UI_path + 'err2.ui', QDialog(self))
        self.uifluerr2.label.setText('Please check parameters to fit')
        self.uifluerr2.fluErrorProgress.setValue(0)
        # self.fluErroFit = myThread()
        # self.fluErrorFit.started.connect()
        # self.fluErrorFit.finished.connect(self.fluErrorResult)
        self.uifluerr2.fluErrorProgress.setMaximum(len(self.fluerr_fit_range))
        self.uifluerr2.cancelPB.clicked.connect(self.uifluerr2.close)
        self.uifluerr2.nextPB.clicked.connect(self.fluErrorFit)

        self.uifluerr2.show()

    def fluErrorFitSingleCore(self, q, i, flu_par,flucal_par,data_to_fit):
        fluerr_result = lm.minimize(fl.flu2min, flu_par,
                                    args=((self.sh, self.qz), flucal_par),
                                    kws={'data': data_to_fit[:, 1], 'eps': data_to_fit[:, 2]})
        q.put([i, fluerr_result.nfree, fluerr_result.redchi])



    def fluErrorFit(self):
        self.uifluerr2.label.setText('Calculating the uncertainty for ' + self.fluerr_pname[0])
        self.uifluerr2.nextPB.setEnabled(False) # unable the next push button
        # create a Parameter() object for fitting
        self.updatePar()  # update parameters to fit with GUI

        # make a copy of parameter just for uncertainty calculation
        fluerr_pname = self.fluerr_pname[0]
        fluerr_par = self.flu_par
        fluerr_cal_par = self.flucal_par
        fluerr_par[fluerr_pname].vary = False # make sure the chosen parameter does not vary

        # time the calculation
        start_time = time.time()
        # self.ErrorFit = myThread(self.fitFlu,errorbar=True,args={})



        # self.fitFlu(uncertainty_calculation=True)
        # fluErrorFitSingleCore_i = partial(fluErrorFitSingleCore2,
        #                                   value_list = self.fluerr_fit_range,
        #                                   sh = self.sh,
        #                                   qz = self.qz,
        #                                   pname = fluerr_pname,
        #                                   flu_par = fluerr_par,
        #                                   flucal_par = fluerr_cal_par,
        #                                   data_to_fit = self.data_to_fit)
        #
        #
        # results = multiCore(fluErrorFitSingleCore_i, range(len(self.fluerr_fit_range)))
        # for pp in results: print(pp)
        # print("Uncertainty calculation takes:", time.time() - start_time, "seconds")
        # for result in results:
        #     self.fluerr_chisq_list[result[0]] = result[3]
        # self.fluerr_nfree = results[-1][2]
        # print(self.fluerr_chisq_list)


        self.fitFlu(uncertainty_calculation=True)
        processes = []  # create a pool for processes
        q = multiprocessing.Queue()
        for i,value in enumerate(self.fluerr_fit_range):
            fluerr_par[fluerr_pname].value = value # change value of the chosen parameter
            p = multiprocessing.Process(target=self.fluErrorFitSingleCore,
                                        args=(q, i, fluerr_par,fluerr_cal_par,self.data_to_fit))
            processes.append(p)
            p.start()
        for process in processes:
            process.join()
        results = [q.get(True) for process in processes]

        for pp in results: print(pp)
        print("Uncertainty calculation takes:", time.time()-start_time, "seconds")
        for result in results:
            self.fluerr_chisq_list[result[0]] = result[2]
        self.fluerr_nfree = results[-1][1]
        print(self.fluerr_chisq_list)

        # # fit data and calculate chisq at each grid point
        # for i,para_value in enumerate(self.fluerr_fit_range):
        #     self.fluerr_parameters[self.fluerr_pname_to_fit].value = para_value
        #     fluresult=lm.minimize(self.flu2min, self.fluerr_parameters, args=(x,y,yerr))
        #     self.fluerr_chisq_list[i] = fluresult.redchi
        #     # update progress
        #
        # self.progressDialog.hide()

        self.uifluerr2.close()
        self.fluErrorResult()

    def fluErrorResult(self):
        # calculate the left/right error for the parameter
        funChisqFactor=interp1d(self.errorlist[:,0],self.errorlist[:,1],kind='cubic')
        chisq_factor = funChisqFactor(self.fluerr_nfree) # chisq_factor corresponding to degree of freedom
        idx_min_chisq = np.argmin(self.fluerr_chisq_list[1:]) + 1
        min_chisq = np.min(self.fluerr_chisq_list[1:])
        self.target_chisq = min_chisq * chisq_factor
        try: # interpolate function of left values against various chisq's
            funChisqListLeft = interp1d(self.fluerr_chisq_list[1:idx_min_chisq+1],
                                        self.fluerr_fit_range[1:idx_min_chisq+1],
                                        kind='linear')
            left_err = self.fluerr_best_value - funChisqListLeft(self.target_chisq)
            left_err_str = format(float(left_err),'.2e')
        except:
            left_err_str = "not found"
        try: # interpolate function of right values against various chisq's
            funChisqListRight = interp1d(self.fluerr_chisq_list[idx_min_chisq:],
                                         self.fluerr_fit_range[idx_min_chisq:],
                                         kind='linear')
            right_err = funChisqListRight(self.target_chisq) - self.fluerr_best_value
            right_err_str = format(float(right_err),'.2e')
        except:
            right_err_str = "not found"

        self.uifluerr3=uic.loadUi(UI_path + 'err3.ui',QDialog(self))
        self.uifluerr3.label.setText('Plot for Chi-square vs Parameter:'+self.fluerr_pname[0])
        self.uifluerr3.minchiLE.setText(format(min_chisq,'.2f'))
        self.uifluerr3.tarchiLE.setText(format(self.target_chisq,'.2f'))
        self.uifluerr3.lefterrLE.setText(left_err_str)
        self.uifluerr3.righterrLE.setText(right_err_str)
        self.uifluerr3.logyCB.stateChanged.connect(self.fluErrorPlot)
        self.uifluerr3.closePB.clicked.connect(lambda x: self.uifluerr3.close())
        self.uifluerr3.closePB.clicked.connect(self.uifluerr3.close)
        self.uifluerr3.savePB.clicked.connect(self.fluErrorSave)
        self.uifluerr3.show()
        self.fluErrorPlot()

    def fluErrorPlot(self):
        the_ax = self.uifluerr3.plotWidget.canvas.ax
        the_ax.clear()
        the_ax.set_xlabel(self.fluerr_pname[0])
        the_ax.set_ylabel('Chi-square')
        # check if y axis is logscale
        if self.uifluerr3.logyCB.checkState()!=0:
            the_ax.set_yscale('log')
        else:
            the_ax.set_yscale('linear')

        # plot the calculated chisq
        the_ax.plot(self.fluerr_fit_range[1:], self.fluerr_chisq_list[1:],
                    marker='o',ls='-')

        # plot the fitted parameter value and corresponding chisq
        the_ax.plot(self.fluerr_fit_range[0], self.fluerr_chisq_list[0],
                    marker='o',color='red')

        # plot the target chisq
        the_ax.plot(self.fluerr_fit_range[[1,-1]],
                    self.target_chisq * np.array([1,1]),
                    ls='-',color='green')

        self.uifluerr3.plotWidget.canvas.draw()

    def fluErrorSave(self):
        print("Save function to be released...")

    def insFluIon(self):  # add one ion in the subphase
        insrows=self.ui.flusubTW.selectionModel().selectedRows()
        insrows=[self.ui.flusubTW.row(self.ui.flusubTW.itemFromIndex(insrows[i])) for i in range(len(insrows))]
        if len(insrows)!=1:
            self.messageBox('Warning:: Only one row can be seleted!')
        else:
            self.ui.flusubTW.insertRow(insrows[0])
        for i in range(3):
            self.ui.flusubTW.setItem(insrows[0],i,QTableWidgetItem('Cl/2/1.80'.split('/')[i]))

    def rmFluIon(self): #remove one ion in the subphase
        rmrows=self.ui.flusubTW.selectionModel().selectedRows()
        removerows=[]
        for rmrow in rmrows:
            removerows.append(self.ui.flusubTW.row(self.ui.flusubTW.itemFromIndex(rmrow)))
            removerows.sort(reverse=True)
        if len(removerows)==0:
            self.messageBox('Warning:: No ion is selected!!')
        else:
            for i in range(len(removerows)):
                self.ui.flusubTW.removeRow(removerows[i])



# from PyQt4 import QtGui

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import matplotlib
matplotlib.use("Qt4Agg")
from matplotlib.backends.backend_qt5agg \
 import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT as NavigationToolbar

class MplCanvas(FigureCanvas):
	def __init__(self):
		self.fig = Figure()
		self.ax = self.fig.add_subplot(111)
		FigureCanvas.__init__(self, self.fig)
		FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
		FigureCanvas.updateGeometry(self)

class MplWidget(QWidget):
	def __init__(self, parent = None):
		QWidget.__init__(self, parent)
		self.canvas = MplCanvas()
		self.toolbar = NavigationToolbar(self.canvas, parent)
		self.vbl = QVBoxLayout()
		self.vbl.addWidget(self.canvas)
		self.vbl.addWidget(self.toolbar)
		self.setLayout(self.vbl)




import sys
import os

#######################################################
# define an exception hook to prevent the app from crashing on exception
# reference: https://stackoverflow.com/questions/38020020/pyqt5-app-exits-on-error-where-pyqt4-app-would-not
sys._excepthook = sys.excepthook
def exception_hook(exctype, value, traceback):
    print(exctype, value, traceback)
    sys._excepthook(exctype, value, traceback)
sys.excepthook = exception_hook

# Use absolute path instead of relative path ('./') to avoid trouble when installed by pip
dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(dir_path)

from PyQt5.QtCore import pyqtRemoveInputHook
from PyQt5.QtWidgets import QApplication
from mainwindow import MainWindow

def main():
    # create application
    pyqtRemoveInputHook()
    app = QApplication(sys.argv)
    app.setApplicationName('15ID-C XFNTR Analyzer')

    # create widget
    w = MainWindow()
    w.setWindowTitle('15ID-C XFNTR Analyzer')
    # w.setWindowIcon(QIcon('logo.png'))
    w.show()

    # connection
    app.lastWindowClosed.connect(app.quit)

    # execute application
    sys.exit(app.exec_())

if __name__ == '__main__':

    main()




from PyQt4 import uic
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtWebKit import *
import numpy as np
import time
import matplotlib as mpl


def cancelErrCal():
    pass
    
def nextErrCal():
    pass

class uiErr1(QDialog):
    def __init__(self, para, num_interval=20): 
        self.iniUI()
        
    def iniUI(self):
        self.ui = uic.loadUi('err1.ui')
        print self.ui   
        self.ui.bestvalLE.setText(format(para.value,'.2e'))
        self.ui.leftranLE.setText(format(para.value*0.9, '.2e')) # left range: 0.1*value
        self.ui.rightranLE.setText(format(para.value*1.1, '.2e')) # right range: 0.1*value
        self.ui.numIntervalLE.setText(format(num_interval+1),'i') # number of intervals for the sample

        self.ui.CancelPB.clicked.connect(self.cancelErrCal)
        self.ui.NextPB.clicked.connect(self.nextErrCal)
        
        
    def paraFitRange(self,para,num_samples):

        best_value, left_limit, right_limit, num_samples = tuple(
            float(ui_err1.bestvalLE.text()),
            float(ui_err1.leftLimitLE.text()),
            float(ui_err1.rightLimitLE.text()),
            float(ui_err1.num_samples.text())
        )
        fit_range = np.linspace(lift_limit,right_limit,num_samples)


from collections import OrderedDict
import periodictable as pdtb
import lmfit as lm
import multiprocessing

r_e = pdtb.constants.electron_radius * 1e10  # classical electron radius, in A
N_A = pdtb.constants.avogadro_number  # Avogadro number, unitless
k_B = 1.38065e-23  # Boltzman constant, in J/K
p_igMu = (1/1.717)*1e-7    # absorption coefficient of 20keV beam by glass
p_thick = 3.5 * 1e7     # thickness of glass tray in A

# define an function that will be used in the program a lot.
absorb = lambda x: np.nan_to_num(np.exp(x))

import numpy as np
import scipy.stats as stat
import fit_ref as mfit
import flu_geometry_routines as gm



def penetrate(beta, delta, alpha, k0):
    alpha[alpha == np.inf] = 0
    alpha = alpha.astype(complex)
    beta_top, beta_bot = beta
    delta_top, delta_bot = delta
    alpha_c = np.sqrt(2 * (delta_bot - delta_top))
    trans = 4 * np.abs(alpha / (alpha + np.sqrt(alpha ** 2 - alpha_c ** 2))) ** 2
    penetration_coeff = 2 * k0 * np.imag(np.sqrt(alpha ** 2 - alpha_c ** 2 + beta_bot * 2j))
    return 1/penetration_coeff, trans

def update_flu_parameters(p, *args):

    assert type(p) is OrderedDict
    # *args have to be at least fitting parameters
    flu_par = args[0]

    # update the fitting parameter whatsoever
    try:
        p['hisc'] = flu_par['hisc'].value  # scale factor for the top phase, unitless.
        p['losc'] = flu_par['losc'].value  # scale factor for the bottom phase, unitless.
        p['bg'] = flu_par['bg'].value  # background intensity, unitless.
        p['tC'] = flu_par['upbk'].value  # ion concentration of the top phase, in M.
        p['bC'] = flu_par['lobk'].value  # ion concentration of the bottom phase, in M.
        p['sC'] = flu_par['surd'].value  # ion surface number density, in A^-2.
        p['qoff'] = flu_par['qoff'].value  # q off set for the data
        p['soff'] = flu_par['soff'].value * 1e7  # det range offset for the measurement
        p['l2off'] = flu_par['loff'].value * 1e7  # l2 offset for the measurement
        p['curv'] = flu_par['curv'].value * 1e10  # the curvature of the interface, in A.
        if p['curv'] == 0: p['curv'] = 10000 * 1e10
    except KeyError as e:
        print("Please check your parameter:{}".format(e))

    if len(args) == 1:
        return p

    # if the *args is tuple (flu_par, sys_par, flu_elements), do the following
    try:
        sys_par = args[1]
        flu_elements = args[2]
    except IndexError:
        print("update_flu_parameters takes 3 extra arguments!")

    # parameterize beam profile
    width = sys_par['width'] * 1e7  # width or FWHM of the beam, in A
    beam_profile = sys_par['beam']
    steps = 500
    if beam_profile == 'Uniform':
        beam_size = width
        weights = np.ones(steps + 1)
    elif beam_profile == 'Gaussian':
        stdev = width / 2.355  # FWHM of the beam, i.e. 2.355 sigma
        beam_size = 2 * (3 * stdev)  # keep the beam up to +/-3 standard deviation, or 99.73% of intensity.
        rays = np.linspace(-beam_size/2, beam_size/2, steps+1)  # devide the beam into 500 rays
        weights = stat.norm(0, stdev).pdf(rays) * width  # weight normalized to the total intensity
    else:
        print("Error: please choose the right beam profile: uniform/gaussian")
        return None

    # unwrap system parameters
    E_inc = float(sys_par['E_inc'])  # energy of incidence, in KeV
    E_emit = float(sys_par['E_emt'])  # energy of  emission, in KeV
    mu_top_inc = float(sys_par['mu_top_inc'] / 1e8)  # abs. coef. of top phase for incidence, in 1/A
    mu_top_emit = float(sys_par['mu_top_emt'] / 1e8)  # abs. coef. of top phase for emission, in 1/A
    mu_bot_inc = float(sys_par['mu_bot_inc'] / 1e8)  # abs. coef. of bot phase for incidence, in 1/A
    mu_bot_emit = float(sys_par['mu_bot_emt'] / 1e8)  # abs. coef. of bot phase for emission, in 1/A
    rho_top = sys_par['rho_top']  # electron density of top pahse, in A^-3
    rho_bot = sys_par['rho_bot']  # electron density of top pahse, in A^-3

    det_len = sys_par['det_len'] * 1e7  # detector length, in A
    span = sys_par['span'] * 1e7

    # calculate abosrption parameters.
    k0 = 2 * np.pi * E_inc / 12.3984  # wave vector for incidence, in A^-1
    k1 = 2 * np.pi * E_emit / 12.3984  # wave vector for emission, in A^-1

    # construct elemental parameters
    vol_top, vol_bot = 0, 0  # vol_bot for ions for 1 L subphase
    ne_top, ne_bot = 0, 0  # total number of electrons from 1 L subphase
    bet_top_inc_ele, bet_top_emit_ele = 0, 0  # beta: for top phase at inc.& emit.
    bet_bot_inc_ele, bet_bot_emit_ele = 0, 0  # beta: for bot phase at inc.& emit.
    flupara_ele = {}  # setup dict for all elements in the subphase
    for i, e in enumerate(flu_elements):
        flupara_ele[i] = \
            [e[0], float(e[1]), float(e[2]),  # name, composition, ionic radius
             pdtb.elements.symbol(e[0]).number,
             pdtb.elements.symbol(e[0]).xray.scattering_factors(energy=sys_par['E_inc'])[1],
             pdtb.elements.symbol(e[0]).xray.scattering_factors(energy=sys_par['E_emt'])[1]]
    for i, p_ in flupara_ele.items():
        n_top_density = p['tC'] * p_[1] * N_A / 1e27  # atoms per A^3 in top phase
        n_bot_density = p['bC'] * p_[1] * N_A / 1e27  # atoms per A^3 in bot phase
        vol_top += n_top_density * 4 / 3 * np.pi * p_[2] ** 3
        vol_bot += n_bot_density * 4 / 3 * np.pi * p_[2] ** 3
        ne_top += n_top_density * p_[3]  # electrons per A^3
        ne_bot += n_bot_density * p_[3]  # electrons per A^3
        bet_top_inc_ele += n_top_density * 2 * np.pi * r_e * p_[4] / k0 ** 2
        bet_top_emit_ele += n_top_density * 2 * np.pi * r_e * p_[5] / k1 ** 2
        bet_bot_inc_ele += n_bot_density * 2 * np.pi * r_e * p_[4] / k0 ** 2
        bet_bot_emit_ele += n_bot_density * 2 * np.pi * r_e * p_[5] / k1 ** 2
    # absorption coefficient and electron density modified by solvent.
    rho_top = ne_top + (1 - vol_top) * rho_top
    rho_bot = ne_bot + (1 - vol_bot) * rho_bot

    # re-evaluate mu
    mu_top_inc = 2 * k0 * bet_top_inc_ele + (1 - vol_top) * mu_top_inc
    mu_top_emit = 2 * k1 * bet_top_emit_ele + (1 - vol_top) * mu_top_emit
    mu_bot_inc = 2 * k0 * bet_bot_inc_ele + (1 - vol_bot) * mu_bot_inc
    mu_bot_emit = 2 * k1 * bet_bot_emit_ele + (1 - vol_bot) * mu_bot_emit

    # calculate beta
    bet_top_inc = mu_top_inc / k0 / 2
    bet_top_emit = mu_top_emit / k1 / 2
    bet_bot_inc = mu_bot_inc / k0 / 2
    bet_bot_emit = mu_bot_emit / k1 / 2

    # calculate delta
    del_top_inc = 2 * np.pi * r_e * rho_top / k0 ** 2  # del=2*PI*re*rho/k^2, unitless #self.flutopdel
    del_top_emit = 2 * np.pi * r_e * rho_top / k1 ** 2
    del_bot_inc = 2 * np.pi * r_e * rho_bot / k0 ** 2
    del_bot_emit = 2 * np.pi * r_e * rho_bot / k1 ** 2

    p['k0'] = k0 # incident ray Energy, in KeV.
    p['k1'] = k1 # emission ray energy, in KeV.
    p['detR'] = det_len # detector range, in mm.
    p['wt'] = weights  # weights for the beam profile, the length of which is the total amount of steps.
    p['bmsz'] = beam_size  # the size of the beam for footprint calculation, in A.
    p['tRho'] = rho_top # electron density of top phase, in A^-3.
    p['bRho'] = rho_bot # electron density of bottom phase, in A^-3.
    p['itMu'] = mu_top_inc # mu for incident beam in top phase, in cm^-1
    p['etMu'] = mu_top_emit # mu for emitted beam in top phase, in cm^-1
    p['ibMu'] = mu_bot_inc # mu for incident beam in bottom phase, in cm^-1
    p['ebMu'] = mu_bot_emit # mu for emitted beam in bottom phase, in cm^-1
    p['itBt'] = bet_top_inc # beta for incident beam in top phase, in cm^-1
    p['etBt'] = bet_top_emit # beta for emitted beam in top phase, in cm^-1
    p['ibBt'] = bet_bot_inc # beta for incident beam in bottom phase, in cm^-1
    p['ebBt'] = bet_bot_emit # beta for emitted beam in bottom phase, in cm^-1
    p['itDt'] = del_top_inc # delta for incident beam in top phase, in cm^-1
    p['etDt'] = del_top_emit # delta for emitted beam in top phase, in cm^-1
    p['ibDt'] = del_bot_inc # delta for incident beam in bottom phase, in cm^-1
    p['ebDt'] = del_bot_emit # delta for emitted beam in bottom phase, in cm^-1
    p['span'] = span # the length of the sample cell, "the span", in A.

    return p

def fluCalFun_core(a0,sh,p):
    
    '''takes in flupara_fit, qz, return fluorescence data.
       Note that 'weights' contains the info of the steps for integration
       a0: the incident angle of X-ray beam, corrected with Qz_offset, in rad.
       sh: the sample height shift w.r.t. its norminal position, in A.
       p['wt']: weights for the beam profile, the length of which is the total amount of steps.
       p['k0']: wavevector for incident ray Energy, in KeV.
       p['detR']: detector range, in mm.
       p['hisc']: scale factor for the top phase, unitless.
       p['losc']: scale factor for the bottom pahse, unitless.
       p['bg']: background intensity, unitless.
       p['k1']: wavevector for emission ray energy, in KeV.
       p['tC']: ion concentration of the top phase, in M.
       p['bC']: ion concentration of the bottom phase, in M.
       p['sC']: ioin surface number density, in A^-2.
       p['qoff']: q off set for the data
       p['soff']: sample height offset (effoct of detector offset included) for the measurement
       p['l2off']: l2 offset for the measurement
       p['tRho']: electron density of top phase, in A^-3.
       p['bRho']: electron density of bottom phase, in A^-3.
       p['itMu']: mu for incident beam in top pahse, in A^-1
       p['etMu']: mu for emitted beam in top phass, in A^-1
       p['ibMu']: mu for incident beam in bottom pahse, in A^-1
       p['ebMu']: mu for emitted beam in bottom phass, in A^-1
       p['itBt']: beta for incident beam in top pahse, in A^-1
       p['etBt']: beta for emitted beam in top phass, in A^-1
       p['ibBt']: beta for incident beam in bottom pahse, in A^-1
       p['ebBt']: beta for emitted beam in bottom phass, in A^-1
       p['itDt']: delta for incident beam in top pahse, in A^-1
       p['etDt']: delta for emitted beam in top phass, in A^-1
       p['ibDt']: delta for incident beam in bottom pahse, in cm^-1
       p['ebDt']: delta for emitted beam in bottom phass, in cm^-1
       p['span']: the length of the sample cell, "the span", in A.
       p['curv']: the curvature of the interface, in A.
       p['bmsz']: the size of the beam for footprint calculation, in A.
       '''


    steps = len(p['wt']) - 1
    fprint = p['bmsz'] / np.sin(a0) # foortprint of the beam on the interface.
    stepsize = fprint / steps
    center = - sh / a0

    # initialize fluorescence data, rows: total, aqueous, organic, interface


    # get the position of single ray hitting the surface
    x0 = np.linspace(center-fprint/2, center+fprint/2, steps+1)
    surface = np.array([gm.hit_surface([xx, 0], -a0, p['curv'], p['span']) for xx in x0])
    hit = np.isfinite(surface[:,0]) * np.isfinite(surface[:,1])  # rays that hit the liquid-liquid interface
    block = np.isfinite(surface[:,0]) * np.isinf(surface[:,1])  # rays that are blocked by the tray
    x_s = surface[:, 0][hit]  # x' for rays that hit on the interface
    z_s = surface[:, 1][hit] # z' for rays that hit on the interface
    wt_s = p['wt'][hit]  # weight for rays that hit on the interface

    # (x,z) and other surface geometry for points where beam hit at the interface.
    theta = -x_s / p['curv']  # incident angle
    a_new = a0 + theta  # actual incident angle w.r. to the surface
    # a1 = a0 + 2 * theta
    a1 = a0
    x_inc = x_s + z_s / a0  # x position where the inc. xray passes z=0 line
    x_ref = x_s - z_s / a1  # x position where the ref. xray passes z=0 line.

    mu_eff_inc = p['etMu'] + p['itMu'] / a0  # eff.abs.depth for incident beam in oil phase
    mu_eff_ref = p['etMu'] - p['itMu'] / a1  # eff.abs.depth for reflected beam in water phase

    # z coordinate of the intersection of ray with following:
    z_inc_l = (x_inc + p['detR'] / 2) * a0  # incidence with left det. boundary: x=-l/2
    z_inc_r = (x_inc - p['detR'] / 2) * a0  # incidence with right det. boundary: x=l/2
    z_ref_l = -(x_ref + p['detR'] / 2) * a1  # reflection with left det. boundary: x=-l/2
    z_ref_r = -(x_ref - p['detR'] / 2) * a1  # reflection with right det. boundary: x=l/2

    # two regions: region3: [-h/2a0,-l/2] & region 2: [-l/2,l/2]
    x_region = [(x_s <= -p['detR'] / 2), (x_s > -p['detR'] / 2) * (x_s < p['detR'] / 2)]

    ################### for redgion 1: region x>= l/2  ########################
    x0_region1 = x0[surface[:, 0] > p['detR'] / 2]  # choose x0 with x'>l/2
    wt_region1 = p['wt'][surface[:, 0] > p['detR'] / 2]  # choose weight with x'>l/2
    upper_bulk1 = wt_region1 * \
                  absorb(-x0_region1 * p['itMu']) / mu_eff_inc * \
                  (absorb((x0_region1 + p['detR'] / 2) * a0 * mu_eff_inc) -
                   absorb((x0_region1 - p['detR'] / 2) * a0 * mu_eff_inc))

    # define the initial intensity to be just background value
    flu = np.array([p['bg']] * 6)

    # if beam miss the surface entirely, do the following:
    if len(x_s) == 0:  # the entire beam miss the interface, only incidence in upper phase.
        # sh_offset_factor = absorb(-mu_top_emit * center[i] * a0)
        usum_inc = stepsize * np.sum(upper_bulk1)
        flu[3] += p['hisc'] * usum_inc * N_A * p['tC'] / 1e27  # oil phase incidence only + background
        flu[0] = flu[3]  # total intensity only contains oil phase
        return flu

    ref = mfit.refCalFun([], [p['tRho'], p['bRho']], [p['itMu'], p['ibMu']], [3.0], 2 * p['k0'] * a_new)
    p_depth, trans = penetrate((p['itBt'],p['ibBt']), (p['itDt'],p['ibDt']), a_new, p['k0'])
    p_depth_eff = 1 / (p['ebMu'] + a_new/a0 / p_depth)

    ################### for region -l/2 < x < l/2  #################
    lower_bulk2 = x_region[1] * wt_s * absorb(-x_s * p['itMu'] - z_s / p_depth) * trans * p_depth * \
                  (absorb(z_s / p_depth_eff) - absorb(z_inc_r / p_depth_eff))
    interface = x_region[1] * wt_s * trans * absorb(-p['itMu'] * x_s)
    upper_bulk2_inc = x_region[1] * wt_s * \
                      (absorb(-x_inc * p['itMu']) / mu_eff_inc * (
                              absorb(z_inc_l * mu_eff_inc) - absorb(z_s * mu_eff_inc)))
    upper_bulk2_inc[np.isnan(upper_bulk2_inc)] = 0  # if there is nan, set to 0
    upper_bulk2_ref = x_region[1] * wt_s * \
                      (absorb(-x_ref * p['itMu']) / mu_eff_ref * ref * (
                              absorb(z_ref_r * mu_eff_ref) - absorb(z_s * mu_eff_ref)))
    upper_bulk2_ref[np.isnan(upper_bulk2_ref)] = 0  # if there is nan, set to 0

    ###################### for region x<=-l/2 ########################
    lower_bulk3 = x_region[0] * wt_s * absorb(-x_s * p['itMu'] - z_s / p_depth) * trans * p_depth_eff * \
                  (absorb(z_inc_l / p_depth_eff) - absorb(z_inc_r / p_depth_eff))
    upper_bulk3 = x_region[0] * wt_s * absorb(-x_ref * p['itMu']) / mu_eff_ref * ref * \
                  (absorb(mu_eff_ref * z_ref_r) - absorb(mu_eff_ref * z_ref_l))
    # if there are rays that are blocked by tray, their intensity is still significent
    if np.sum(block) > 0:
        edge = None
    # combine the two regions and integrate along x direction by performing np.sum.
    bsum = stepsize * np.sum(lower_bulk3 + lower_bulk2)
    ssum = stepsize * np.sum(interface)
    usum_inc = stepsize * (np.sum(upper_bulk1) + np.sum(upper_bulk2_inc))
    usum_ref = stepsize * (np.sum(upper_bulk3) + np.sum(upper_bulk2_ref))


    # vectorized integration method is proved to reduce the computation time by a factor of 5 to 10.
    int_bulk = p['losc'] * bsum * N_A * p['bC'] / 1e27
    int_upbk_inc = p['hisc'] * usum_inc * N_A * p['tC'] / 1e27  # metal ions in the upper phase.
    int_upbk_ref = p['hisc'] * usum_ref * N_A * p['tC'] / 1e27  # metal ions in the upper phase.
    int_sur = p['losc'] * ssum * p['sC']

    flu += np.array([int_bulk+int_sur+int_upbk_inc+int_upbk_ref,  # 2. total fluorescence + background
                     int_bulk,  # 3. lower bulk
                     int_sur,  # 4. interface
                     int_upbk_inc+int_upbk_ref, # 5. upper bulk
                     int_upbk_inc,  # 6. upper bulk incidence
                     int_upbk_ref])  # 7. upper bulk reflection
    return flu

def flu2min(pars, x, p, data=None, eps=None): # residuel for flu fitting

    sh, qz = x
    p = update_flu_parameters(p, (pars))

    alpha = (qz + p['qoff']) / p['k0'] / 2 # include the qz offset.
    a0006 = 0.006 / p['k0'] / 2  # incident angle for qz=0.006

    # initialize fluorescence data 3-D matrix
    flu = np.zeros((len(sh), len(qz), 8))
    try:
        for i, ds in enumerate(sh):
            for j, a0 in enumerate(alpha):
                dsh = -p['l2off'] * (a0 - a0006) + p['soff'] + ds*1e7
                flu[i, j, 0] = ds
                flu[i, j, 1] = qz[j]
                flu[i, j, 2:] = fluCalFun_core(a0, dsh, p)
    except KeyError as e:
        print("Please check parameter: {}".format(e))
    if data is None:
        return flu
    if eps is None:
        return (flu[:,:,2].flatten() - data)

    return (flu[:,:,2].flatten() - data) / eps

def fluErrorFitSingleCore2(i, value_list, sh, qz, pname, flu_par, flucal_par, data_to_fit):
    flu_par[pname].value = value_list[i]  # change value of the chosen parameter
    fluerr_result = lm.minimize(fl.flu2min, flu_par,
                                args=((sh, qz), flucal_par),
                                kws={'data': data_to_fit[:, 1], 'eps': data_to_fit[:, 2]})
    return ([i, value_list[i], fluerr_result.nfree, fluerr_result.redchi])

def multiCore(func, iterable):
    pool = multiprocessing.Pool()
    result = pool.map(func, range(len(iterable)))
    pool.close()
    pool.join()
    return result

if __name__ == '__main__':
    p_depth, trans = penetrate((1.35e-10, 5.0e-10), (4.466e-7, 5.84e-7), np.array([3.14]), 1.1355)
    print(p_depth)
    print(trans)



# All .ui files and .so files are added through keyword: package_data, because setuptools doesn't include them automatically.
import sys
import os
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = "xfntr",
    version = "0.2.3",
    author = "Zhu Liang",
    author_email = "zliang8@uic.edu",
    description = "A software that analyzes xfntr data",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/zhul9311/XFNTR.git",
    packages = find_packages(),
    package_dir = {'':'.'},
    package_data = {
        '' : ['xr_ref.cpython-37m-darwin.so',
              'GUI/*',
              'images/*',
              'test/*']
    },
    exclude_package_data = {
        '' : ['.git/','.setup.py.swp']
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires = '>=3.6',
    install_requires = [
        'pyqt5',
        'scipy',
        'matplotlib',
        'lmfit',
        'periodictable'
        'numba'
    ],
    entry_points = { # create scripts and add to sys.PATH
        'console_scripts':[
            'xfntr1 = xfntr.main:main'
        ],
        'gui_scripts': [
            'xfntr = xfntr.main:main'
        ]
    },
)



import numpy as np
import matplotlib.pyplot as plt

def laser_gun(origin, alpha, wall=None,celling=None):
    '''
        Define a laser starting from a point with a certain angle,
    until it hits a wall/celling. 
        Input: origin: np.array([x,y])
        Returns the x/y of the hitting point.
    '''
    if wall != None:
        x = wall
        y = origin[1] + np.tan(alpha) * (x - origin[0])
    elif celling != None:
        y = celling
        x = origin[0] + 1/np.tan(alpha) * (celling - origin[1])
    else:
        print("please define a stop")
        return None
    return np.array([x, y])


def central_dist(point, line):
    ''' 
    Calculate the distance from a point to a line.
    Line is defined as: A*x + B*y + C = 0.
    Input line takes in (A,B,C), point takes in [x,y]
    '''
    x, y = point[0], point[1]
    A, B, C = line
    d = np.abs(A*x+B*y+C)/np.sqrt(A**2+B**2)
    return d


def surf_points(x,R):
    '''return an array of surface points given x coordinates
       Input: x, 1D array; curvature in m '''
    x = np.array(x)
    theta = -x / R # sin(theta) = theta, no difference
    y = - R * theta**2/2  # cos(theta) = 1 - theta**2/2, no difference
    surface = np.array([x,y])
    return np.transpose(surface), theta

def hit_surface(point, alpha, curvature, span, sh=0):
    '''
        Given a point and angle, calculate its laser interactig with the surface
        input: points, 1-D array
        output: points, 2-D array
        source: http://mathworld.wolfram.com/Circle-LineIntersection.html. 
    Note that this algorithm applies to circle center at (0,0), sh is the sample height relative NOM position.
        
    '''
    # if point[0]==0.: return np.array([0,0])
    R = curvature 
    cell_left = - span / 2
    cell_right = span / 2
    pt1 = np.array([point[0],point[1]+(R-sh)]) # raise the line up by R.
    pt2 = laser_gun(pt1,alpha,wall=-50) # second point for the line
    
    dx = pt2[0] - pt1[0]
    dy = pt2[1] - pt1[1]
    dr = np.sqrt(dx**2+dy**2)
    D = pt1[0] * pt2[1] - pt2[0] * pt1[1]
    delta = (R*dr)**2-D**2
    if delta < 0:
        hit = np.array([np.inf,np.inf])
    else:
        x1 = (D*dy + dy/abs(dy)*dx*np.sqrt(delta)) / dr**2
        y1 = (-D*dx + abs(dy)*np.sqrt(delta)) / dr**2
#     x2 = (D*dy + sign(dy)*dx*np.sqrt(delta)) / dr**2
#     y2 = (D*dx + abs(dy)*np.sqrt(delta)) / dr**2
        hit = np.array([x1,y1-(R-sh)]) # put the line down by R-sh.
        if hit[0] > cell_right or hit[0] < cell_left:
            hit = np.array([hit[0], np.inf])
    return hit